declare type IntByteLength = 1 | 2 | 3 | 4;
declare class IRValue {
    name: string;
    symbol: string;
    constval?: IRConst;
    readinsns?: Set<IRInsn>;
}
declare class IRValueState {
    constval?: IRConst;
}
declare class IRConst {
    min: number;
    max: number;
    on: number;
    mask: number;
    constructor(min: number, max: number, on: number, mask: number);
    merge(b: IRConst): IRConst;
    add(b: IRConst): IRConst;
    sub(b: IRConst): IRConst;
    or(b: IRConst): IRConst;
    and(b: IRConst): IRConst;
    xor(b: IRConst): IRConst;
    shl(b: IRConst): IRConst;
    shr(b: IRConst): IRConst;
    sar(b: IRConst): IRConst;
}
declare abstract class IRInsn {
    opcode: string;
    exec(state: IRBlockState): void;
}
declare abstract class IRBlockInsn extends IRInsn {
}
declare abstract class IRValueInsn extends IRBlockInsn {
    assign: IRValue;
}
declare abstract class IRJumpInsn extends IRInsn {
}
declare class IR_JMP extends IRJumpInsn {
    opcode: string;
    target: IRBlock;
}
declare class IR_JNZ extends IRJumpInsn {
    opcode: string;
    test: IRValue;
    iftrue: IRBlock;
    iffalse: IRBlock;
    exec(state: IRBlockState): void;
}
declare class IR_RET extends IRJumpInsn {
    opcode: string;
    value: IRValue;
}
declare type BinopOpcode = "add" | "sub" | "or" | "and" | "xor" | "sar" | "shr" | "shl" | "div" | "mul" | "udiv" | "rem" | "urem";
declare type CompareOpcode = "eq" | "ne" | "sle" | "slt" | "sge" | "sgt" | "ule" | "ult" | "uge" | "ugt";
declare class IRBinopInsn extends IRValueInsn {
    opcode: BinopOpcode;
    left: IRValue;
    right: IRValue;
    exec(state: IRBlockState): void;
}
declare class IRCompareInsn extends IRValueInsn {
    opcode: CompareOpcode;
    left: IRValue;
    right: IRValue;
    exec(state: IRBlockState): void;
}
declare class IR_COPY extends IRValueInsn {
    opcode: string;
    value: IRValue;
    exec(state: IRBlockState): void;
}
declare class IR_LOAD extends IRValueInsn {
    opcode: string;
    bytes: IntByteLength;
    address: IRValue;
    exec(state: IRBlockState): void;
}
declare class IR_STORE extends IRBlockInsn {
    opcode: string;
    bytes: IntByteLength;
    address: IRValue;
    value: IRValue;
}
export declare class IRBlock {
    scope: IRFunctionScope;
    label?: string;
    insns: IRBlockInsn[];
    jump: IRJumpInsn;
    constructor(scope: IRFunctionScope, label?: string);
    append<T extends IRBlockInsn>(insn: T): T;
    finish<T extends IRJumpInsn>(jump: T): T;
    load(assign: IRValue, address: IRValue, bytes: IntByteLength): IR_LOAD;
    store(value: IRValue, address: IRValue, bytes: IntByteLength): IR_STORE;
    copy(assign: IRValue, value: IRValue): IR_COPY;
    binop(op: BinopOpcode, assign: IRValue, left: IRValue, right: IRValue): IRBinopInsn;
    cmp(op: CompareOpcode, assign: IRValue, left: IRValue, right: IRValue): IRCompareInsn;
    ret(value: IRValue): IR_RET;
    jmp(target: IRBlock): IR_JMP;
    jnz(test: IRValue, iftrue: IRBlock, iffalse: IRBlock): IR_JNZ;
    states: Map<IRBlock, IRBlockState>;
    finalstate: IRBlockState;
    analyze(prevblock: IRBlock): void;
    dump(): void;
}
declare class IRBlockState {
    vars: Map<string, IRValueState>;
    readset: Set<string>;
    prev: IRBlock;
    curr: IRBlock;
    constructor(prev: IRBlock, curr: IRBlock);
    read(insn: IRInsn, slot: IRValue): IRValueState;
    write(slot: IRValue, value: IRValueState): void;
    merge(that: IRBlockState): IRBlockState;
}
export declare class IRFunctionScope {
    locals: Map<string, IRValue>;
    blocks: IRBlock[];
    block(label: string): IRBlock;
    temp(name: string): IRValue;
    const16(x: number): IRValue;
    constx(x: number): IRValue;
    symaddr(sym: string): IRValue;
    finish(): void;
    dump(): void;
}
export {};
